﻿local GCD = LibStub("AceAddon-3.0"):NewAddon("GCD", "AceConsole-3.0", "AceEvent-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("GCD")
local ACD = LibStub("AceConfigDialog-3.0")
local media = LibStub("LibSharedMedia-3.0")
GCD.version = 44

local GetTime = GetTime
local sin, cos, rad, ceil, floor = math.sin, math.cos, math.rad, math.ceil, math.floor
local BOOKTYPE_SPELL = BOOKTYPE_SPELL

local GCDBar, startTime, duration, spellNum, warned, castStartTime, castEndTime, castDuration, castInterrupted, castLatency, castSent, isCasting, isGCD

local cdFrames = {}
local frameBuf = {
	["Frame"] = {}
}

local function createFrame(frameType, frameName, frameParent)
	local frame = nil
	if #frameBuf[frameType] > 0 then
		frame = tremove(frameBuf[frameType])
	else
		frame = CreateFrame(frameType, frameName, frameParent)
	end
	return frame
end

local function deleteFrame(frame)
	frame:Hide()
	tinsert(frameBuf[frame:GetObjectType()], frame)
end

local function GetSpellPosInSpellbook(spellName)
	local spellNum = nil
	for tab = 1, 4 do
		local _, _, offset, numSpells = GetSpellTabInfo(tab)
		for i = (1+offset), (offset+numSpells) do
			local spell = GetSpellName(i, BOOKTYPE_SPELL)
			if strlower(spell) == strlower(spellName) then
				spellNum = i
				break
			end
		end
	end
	return spellNum;
end

local cdSpells = {
	["DRUID"] = {
		{['spellID'] = 53251, ['xOff'] = 0, ['yOff'] = 0},
		{['spellID'] = 18562, ['xOff'] = 0, ['yOff'] = 0},
		{['spellID'] = 17116, ['xOff'] = 0, ['yOff'] = 0}
	},
	["PALADIN"] = {
		{['spellID'] = 48825, ['xOff'] = 0, ['yOff'] = 0}
	},
	["PRIEST"] = {
		{['spellID'] = 53007, ['xOff'] = 0, ['yOff'] = 0},
		{['spellID'] = 48113, ['xOff'] = 0, ['yOff'] = 0},
		{['spellID'] = 48089, ['xOff'] = 0, ['yOff'] = 0}
	},
	["SHAMAN"] = {
		{['spellID'] = 61301, ['xOff'] = 0, ['yOff'] = 0}
	}
}

local spells = {
	["DRUID"] = 5185,
	["PALADIN"] = 635,
	["PRIEST"] = 2050,
	["SHAMAN"] = 331,
	["WARRIOR"] = 6673,
	["DEATHKNIGHT"] = 45902,
	["HUNTER"] = 75,
	["MAGE"] = 1459,
	["WARLOCK"] = 687,
	["ROGUE"] = 1752,
}

function GCD:InitializeDefaults()
	local _, class = UnitClass("player")
	self.defaults = {
		char = {
			spark = true,
			spellCds = true,
			castTime = true,
			hideCastBar = false,
			spellText = false,
			ring = true,
			attachToMouse = true,
			cdSpells = cdSpells[class] or {},
			talentSpecs = {
				primary = true,
				secondary = true
			}
		},
		profile = {
			position = {
				x = 400,
				y = 400
			},
			spark = {
				color = {r=1, g=1, b=1, a=1},
				texture = "Spells\\AURA_01"
			},
			ring = {
				color = {r=0, g=1, b=0, a=0.5},
				texture = "Spells\\AURARUNE8",
				rotate = true
			},
			castTime = {
				castColor = {r=0, g=1, b=0, a=1},
				latencyColor = {r=1, g=0, b=0, a=1},
				interruptColor = {r=1, g=1, b=0, a=1},
				texture = "Spells\\AURA_01",
				spellText = {
					font = "Calibri",
					fontSize = 12
				}
			}
		}
	}

	for i,v in ipairs(self.defaults.char.cdSpells) do
		v.xOff = cos(rad(((i - 1) / #self.defaults.char.cdSpells ) * 360)) * 36 + 10
		v.yOff = sin(rad(((i - 1) / #self.defaults.char.cdSpells ) * 360)) * 36 - 10
	end
end

function GCD:InitializeOptions()
	self.options = {
		name = "GCD",
		type = "group",
		args = {
			unlock = {
				name = L["Unlock"],
				type = "execute",
				func = function() GCD:Unlock() end,
				order = 1
			},
			info = {
				name = L["All specified colors are vertex colors! (see wowwiki.com)"],
				type = "description",
				order = 2
			},
			general = {
				name = L["General"],
				type = "group",
				args = {
					position = {
						name = L["Position"],
						type = "header",
						order = 0
					},
					mouse = {
						name = L["Attach to cursor"],
						type = "toggle",
						get = function(info) return self.db.char.attachToMouse end,
						set = function(info,val)
								self.db.char.attachToMouse = val
								self:ApplySettings()
							end,
						order = 1
					},
					x = {
						name = L["Left"],
						type = "input",
						get = function(info) return tostring(self.db.profile.position.x) end,
						set = function(info, val)
									self.db.profile.position.x = tonumber(val)
									self:ApplySettings()
								end,
						validate = function(_, value)
											if tonumber(value) then
												return true
											else
												return L["Left has to be a number!"]
											end
										end,
						order = 2
					},
					y = {
						name = L["Bottom"],
						type = "input",
						get = function(info) return tostring(self.db.profile.position.y) end,
						set = function(info, val)
									self.db.profile.position.y = tonumber(val)
									self:ApplySettings()
								end,
						validate = function(_, value)
											if tonumber(value) then
												return true
											else
												return L["Bottom has to be a number!"]
											end
										end,
						order = 3
					},
					specs = {
						name = L["Talent Specs"],
						type = "header",
						order = 10
					},
					primary = {
						name = L["Primary"],
						type = "toggle",
						get = function(info) return self.db.char.talentSpecs.primary end,
						set = function(info,val)
								self.db.char.talentSpecs.primary = val
								self:ACTIVE_TALENT_GROUP_CHANGED()
							end,
						order = 11
					},
					secondary = {
						name = L["Secondary"],
						type = "toggle",
						get = function(info) return self.db.char.talentSpecs.secondary end,
						set = function(info,val)
								self.db.char.talentSpecs.secondary = val
								self:ACTIVE_TALENT_GROUP_CHANGED()
							end,
						order = 12
					},
				},
				order = 10
			},
			spark = {
				name = L["Spark"],
				type = "group",
				args = {
					general = {
						name = L["General"],
						type = "header",
						order = 0
					},
					enabled = {
						name = L["Enabled"],
						type = "toggle",
						get = function(info) return self.db.char.spark end,
						set = function(info,val)
								self.db.char.spark = val
								self:ApplySettings()
							end,
						order = 1
					},
					display = {
						name = L["Display"],
						type = "header",
						order = 10
					},
					texture = {
						name = L["Texture"],
						type = "input",
						disabled = function() return not self.db.char.spark end,
						get = function(info) return self.db.profile.spark.texture end,
						set = function(info, val)
									self.db.profile.spark.texture = val
									self:ApplySettings()
								end,
						order = 11
					},
					color = {
						name = L["Color"],
						type = "color",
						disabled = function() return not self.db.char.spark end,
						get = function(info) return self.db.profile.spark.color.r, self.db.profile.spark.color.g, self.db.profile.spark.color.b, self.db.profile.spark.color.a end,
						set = function(info, r, g, b, a)
									self.db.profile.spark.color = {r=r, g=g, b=b, a=a}
									self:ApplySettings()
								end,
						hasAlpha = true,
						order = 12
					},
					misc = {
						name = L["Miscellaneous"],
						type = "header",
						order = 20
					},
					defaults = {
						name = L["Restore defaults"],
						type = "execute",
						disabled = function() return not self.db.char.spark end,
						func = function()
									self.db.profile.spark = self.defaults.profile.spark
									self:ApplySettings()
								end,
						order = 21
					},
				},
				order = 12
			},
			ring = {
				name = L["Ring"],
				type = "group",
				args = {
					display = {
						name = L["Display"],
						type = "header",
						order = 0
					},
					enabled = {
						name = L["Enabled"],
						type = "toggle",
						get = function(info) return self.db.char.ring end,
						set = function(info,val)
								self.db.char.ring = val
								self:ApplySettings()
							end,
						order = 1
					},
					texture = {
						name = L["Texture"],
						type = "input",
						disabled = function() return not self.db.char.ring end,
						get = function(info) return self.db.profile.ring.texture end,
						set = function(info, val)
									self.db.profile.ring.texture = val
									self:ApplySettings()
								end,
						order = 2
					},
					color = {
						name = L["Color"],
						type = "color",
						disabled = function() return not self.db.char.ring end,
						get = function(info) return self.db.profile.ring.color.r, self.db.profile.ring.color.g, self.db.profile.ring.color.b, self.db.profile.ring.color.a end,
						set = function(info, r, g, b, a)
									self.db.profile.ring.color = {r=r, g=g, b=b, a=a}
									self:ApplySettings()
								end,
						hasAlpha = true,
						order = 3
					},
					rotate = {
						name = L["Rotate"],
						type = "toggle",
						disabled = function() return not self.db.char.ring end,
						get = function(info) return self.db.profile.ring.rotate end,
						set = function(info, val) self.db.profile.ring.rotate = val end,
						order = 4
					},
					misc = {
						name = L["Miscellaneous"],
						type = "header",
						order = 10
					},
					defaults = {
						name = L["Restore defaults"],
						type = "execute",
						disabled = function() return not self.db.char.ring end,
						func = function()
									self.db.profile.ring = self.defaults.profile.ring
									self:ApplySettings()
								end,
						order = 11
					},
				},
				order = 11
			},
			castTime = {
				name = L["Casttime"],
				type = "group",
				args = {
					general = {
						name = L["General"],
						type = "header",
						order = 0
					},
					enabled = {
						name = L["Enabled"],
						type = "toggle",
						get = function(info) return self.db.char.castTime end,
						set = function(info,val)
								self.db.char.castTime = val
								if val then
									self:RegisterEvent("UNIT_SPELLCAST_SENT")
									self:RegisterEvent("UNIT_SPELLCAST_START")
									self:RegisterEvent("UNIT_SPELLCAST_STOP")
									self:RegisterEvent("UNIT_SPELLCAST_INTERRUPTED")
									self:RegisterEvent("UNIT_SPELLCAST_FAILED")
									self:RegisterEvent("UNIT_SPELLCAST_DELAYED")
									self:RegisterEvent("UNIT_SPELLCAST_CHANNEL_START")
									self:RegisterEvent("UNIT_SPELLCAST_CHANNEL_STOP")
									self:RegisterEvent("UNIT_SPELLCAST_CHANNEL_UPDATE")
								else
									self:UnregisterEvent("UNIT_SPELLCAST_SENT")
									self:UnregisterEvent("UNIT_SPELLCAST_START")
									self:UnregisterEvent("UNIT_SPELLCAST_STOP")
									self:UnregisterEvent("UNIT_SPELLCAST_INTERRUPTED")
									self:UnregisterEvent("UNIT_SPELLCAST_FAILED")
									self:UnregisterEvent("UNIT_SPELLCAST_DELAYED")
									self:UnregisterEvent("UNIT_SPELLCAST_CHANNEL_START")
									self:UnregisterEvent("UNIT_SPELLCAST_CHANNEL_STOP")
									self:UnregisterEvent("UNIT_SPELLCAST_CHANNEL_UPDATE")
								end
								self:ApplySettings()
							end,
						order = 1
					},
					hideCastBar = {
						name = L["Hide default castbar"],
						type = "toggle",
						disabled = function() return not self.db.char.castTime end,
						get = function(info) return self.db.char.hideCastBar end,
						set = function(info,val)
								self.db.char.hideCastBar = val
								if val then
									CastingBarFrame:UnregisterAllEvents()
									CastingBarFrame:Hide()
								else
									CastingBarFrame:UnregisterAllEvents()
									CastingBarFrame:RegisterEvent("UNIT_SPELLCAST_START")
									CastingBarFrame:RegisterEvent("UNIT_SPELLCAST_STOP")
									CastingBarFrame:RegisterEvent("UNIT_SPELLCAST_FAILED")
									CastingBarFrame:RegisterEvent("UNIT_SPELLCAST_INTERRUPTED")
									CastingBarFrame:RegisterEvent("UNIT_SPELLCAST_DELAYED")
									CastingBarFrame:RegisterEvent("UNIT_SPELLCAST_CHANNEL_START")
									CastingBarFrame:RegisterEvent("UNIT_SPELLCAST_CHANNEL_STOP")
									CastingBarFrame:RegisterEvent("UNIT_SPELLCAST_CHANNEL_UPDATE")
								end
							end,
						order = 2
					},
					spark = {
						name = L["Spark"],
						type = "header",
						order = 10
					},
					texture = {
						name = L["Texture"],
						type = "input",
						disabled = function() return not self.db.char.castTime end,
						get = function(info) return self.db.profile.castTime.texture end,
						set = function(info, val)
									self.db.profile.castTime.texture = val
									self:ApplySettings()
								end,
						order = 11
					},
					castColor = {
						name = L["Cast color"],
						type = "color",
						disabled = function() return not self.db.char.castTime end,
						get = function(info) return self.db.profile.castTime.castColor.r, self.db.profile.castTime.castColor.g, self.db.profile.castTime.castColor.b, self.db.profile.castTime.castColor.a end,
						set = function(info, r, g, b, a)
									self.db.profile.castTime.castColor = {r=r, g=g, b=b, a=a}
									self:ApplySettings()
								end,
						hasAlpha = true,
						order = 12
					},
					latencyColor = {
						name = L["Latency color"],
						type = "color",
						disabled = function() return not self.db.char.castTime end,
						get = function(info) return self.db.profile.castTime.latencyColor.r, self.db.profile.castTime.latencyColor.g, self.db.profile.castTime.latencyColor.b, self.db.profile.castTime.latencyColor.a end,
						set = function(info, r, g, b, a)
									self.db.profile.castTime.latencyColor = {r=r, g=g, b=b, a=a}
									self:ApplySettings()
								end,
						hasAlpha = true,
						order = 13
					},
					interruptColor = {
						name = L["Interrupt color"],
						type = "color",
						disabled = function() return not self.db.char.castTime end,
						get = function(info) return self.db.profile.castTime.interruptColor.r, self.db.profile.castTime.interruptColor.g, self.db.profile.castTime.interruptColor.b, self.db.profile.castTime.interruptColor.a end,
						set = function(info, r, g, b, a)
									self.db.profile.castTime.interruptColor = {r=r, g=g, b=b, a=a}
									self:ApplySettings()
								end,
						hasAlpha = true,
						order = 14
					},
					spellText = {
						name = L["Spell Text"],
						type = "header",
						order = 20
					},
					spellTextEnabled = {
						name = L["Enabled"],
						type = "toggle",
						disabled = function() return not self.db.char.castTime end,
						get = function(info) return self.db.char.spellText end,
						set = function(info,val)
								self.db.char.spellText = val
								self:ApplySettings()
							end,
						order = 21
					},
					font = {
						name = L["Font"],
						type = "select",
						disabled = function() return (not self.db.char.castTime or not self.db.char.spellText) end,
						dialogControl = 'LSM30_Font',
						get = function() return self.db.profile.castTime.spellText.font end,
						set = function(_, value)
									self.db.profile.castTime.spellText.font = value
									self:ApplySettings()
								end,
						values = media:HashTable("font"),
						order = 22
					},
					fontSize = {
						name = L["Font Size"],
						type = "range",
						disabled = function() return (not self.db.char.castTime or not self.db.char.spellText) end,
						min = 1,
						max = 30,
						step = 1,
						get = function() return self.db.profile.castTime.spellText.fontSize end,
						set = function(_, value)
									self.db.profile.castTime.spellText.fontSize = value
									self:ApplySettings()
								end,
						order = 23
					},
					misc = {
						name = L["Miscellaneous"],
						type = "header",
						order = 30
					},
					defaults = {
						name = L["Restore defaults"],
						type = "execute",
						disabled = function() return not self.db.char.castTime end,
						func = function()
									self.db.profile.castTime = self.defaults.profile.castTime
									self.db.char.castTime = self.defaults.char.castTime
									self.db.char.castTime = self.defaults.char.spellText
									self:ApplySettings()
								end,
						order = 31
					}
				},
				order = 13
			},
			spellCds = {
				name = L["Spellcds"],
				type = "group",
				childGroups = "tab",
				args = {
					enabled = {
						name = L["Enabled"],
						type = "toggle",
						get = function(info) return self.db.char.spellCds end,
						set = function(info,val)
								self.db.char.spellCds = val
								self:ApplySettings()
								if val then
									self:ACTIONBAR_UPDATE_COOLDOWN()
								else
									for _,v in ipairs(cdFrames) do
										if v.frame then v.frame:Hide() end
									end
								end
							end,
						order = 100
					},
					defaults = {
						name = L["Restore defaults"],
						type = "execute",
						disabled = function() return not self.db.char.spellCds end,
						func = function()
								self.db.char.cdSpells = self.defaults.char.cdSpells
								self:SPELLS_CHANGED()
							end,
						order = 200
					},
					add = {
						name = L["Add Spell"],
						type = "input",
						disabled = function() return not self.db.char.spellCds end,
						get = false,
						set = function(info, value)
									tinsert(self.db.char.cdSpells, {['spellID'] = tonumber(value), ['xOff'] = 0, ['yOff'] = 0})
									self:PopulateCdSpellsOptions()
									self:SPELLS_CHANGED()
							end,
						validate = function(info, value)
								if tonumber(value) and GetSpellPosInSpellbook(GetSpellInfo(tonumber(value))) then
									return true
								else
									return string.format("%s is not a spell u possess!", value)
								end
							end,
						order = 300
					},
					spells = {
						name = L["Spells"],
						type = "group",
						disabled = function() return not self.db.char.spellCds end,
						args = {}
					}
				},
				order = 14
			}
		}
	}
end

function GCD:PopulateCdSpellsOptions()
	for i,v in ipairs(self.db.char.cdSpells) do
		local arg = {
			name = tostring(GetSpellInfo(v.spellID)),
			type = "group",
			args = {
				xPos = {
					name = L["x-Offset"],
					type = "input",
					get = function() return tostring(self.db.char.cdSpells[i].xOff) end,
					set = function(info, value) self.db.char.cdSpells[i].xOff = tonumber(value) end,
					validate = function(info, value) if tonumber(value) == nil then return "That's not a number!" end return true end
				},
				yPos = {
					name = L["y-Offset"],
					type = "input",
					get = function() return tostring(self.db.char.cdSpells[i].yOff) end,
					set = function(info, value) self.db.char.cdSpells[i].yOff = tonumber(value) end,
					validate = function(info, value) if tonumber(value) == nil then return "That's not a number!" end return true end
				},
				remove = {
					name = L["Remove Spell"],
					type = "execute",
					func = function()
								tremove(self.db.char.cdSpells, i)
								self.options.args.spellCds.args.spells.args[tostring(i)] = nil
								self:SPELLS_CHANGED()
							end
				}
			}
		}
		self.options.args.spellCds.args.spells.args[tostring(i)] = arg
	end
end

local function OnUpdate(self, elapsed)
	if not isCasting and not isGCD then
		GCDBar:Hide()
	end

	local x, y
	if GCD.db.char.attachToMouse then
		x, y = GetCursorPosition(UIParent);
	else
		x, y = GCD.db.profile.position.x, GCD.db.profile.position.y
	end
	self:ClearAllPoints()
	self:SetPoint('CENTER', UIParent, 'BOTTOMLEFT', (x+self.xOff)/self:GetEffectiveScale(), (y+self.yOff)/self:GetEffectiveScale())
	
	if self.spark then
		if GCD.db.char.ring and GCD.db.profile.ring.rotate then
			self.texture.timer = self.texture.timer + elapsed;
			if ( self.texture.timer > 0.02 ) then
				self.texture.hAngle = self.texture.hAngle + 0.5;
				local s = sin(rad(self.texture.hAngle));
				local c = cos(rad(self.texture.hAngle));
				self.texture:SetTexCoord(	0.5-s, 0.5+c,
													0.5+c, 0.5+s,
													0.5-c, 0.5-s,
													0.5+s, 0.5-c);
				self.texture.timer = 0;
			end
		end
		if isGCD and GCD.db.char.spark then
			local gcdPerc = (GetTime() - self.startTime) / self.duration
			gcdPerc = gcdPerc < 1 and gcdPerc or 1			
			local angle = rad(90-gcdPerc*360)
			local x, y = cos(angle), sin(angle)
			self.spark:ClearAllPoints()
			self.spark:SetPoint('CENTER', self, 'CENTER',x*25,y*25)
			self.spark:Show()
			if gcdPerc >= 1 then
				isGCD = false
				self.spark:Hide()
			end
		else
			self.spark:Hide()
		end
	end
	if self.castSpark then
		if isCasting and GCD.db.char.castTime then
			local castPerc = (1000 * GetTime() - castStartTime) / castDuration
			castPerc = castPerc < 1 and castPerc or 1
			if castPerc > 1-castLatency then
				self.castSpark:SetVertexColor(GCD.db.profile.castTime.latencyColor.r, GCD.db.profile.castTime.latencyColor.g, GCD.db.profile.castTime.latencyColor.b, GCD.db.profile.castTime.latencyColor.a) -- 1,0,0,1
			elseif castInterrupted then
				self.castSpark:SetVertexColor(GCD.db.profile.castTime.interruptColor.r, GCD.db.profile.castTime.interruptColor.g, GCD.db.profile.castTime.interruptColor.b, GCD.db.profile.castTime.interruptColor.a) -- 1,1,0,1
			else
				self.castSpark:SetVertexColor(GCD.db.profile.castTime.castColor.r, GCD.db.profile.castTime.castColor.g, GCD.db.profile.castTime.castColor.b, GCD.db.profile.castTime.castColor.a) -- 0,1,0,1
			end
			local angle = rad(90-castPerc*360)
			local x, y = cos(angle), sin(angle)
			self.castSpark:ClearAllPoints()
			self.castSpark:SetPoint('CENTER', self, 'CENTER',x*30,y*30)
			self.castSpark:Show()
			if GCD.db.char.spellText then
				self.spellText:Show()
			end
			if castPerc >= 1 then
				isCasting = false
				self.castSpark:Hide()
				if GCD.db.char.spellText then
					self.spellText:Hide()
				end
			end
		else
			self.castSpark:Hide()
			if GCD.db.char.spellText then
				self.spellText:Hide()
			end
		end
	end
	if self.spell and GCD.db.char.spellCds then
		local perc = (GetTime() - self.startTime) / self.duration
		self.texture:SetVertexColor(1.0,perc,perc)
		local dur = floor((self.duration - (GetTime() - self.startTime))*10)/10
		if dur > 99 then
			dur = '>>'
		elseif dur > 10 then
			dur = ceil(dur)
		end
		self.cdText:SetText(dur)
		if perc >= 1 then
			self:Hide()
		end
	end
end

local function OnHide(self)
	self:SetScript('OnUpdate', nil)
end

local function OnShow(self)
	self:SetScript('OnUpdate', OnUpdate)
end

function GCD:FixDatabase()
	if not self.db.char.cdSpells then
		local _, class = UnitClass("player")
		self.db.char.cdSpells = cdSpells[class] or {}
	end
end

function GCD:OnInitialize()
	self:InitializeDefaults()
	self.db = LibStub("AceDB-3.0"):New("GCDDB", self.defaults)
	self:FixDatabase()
	
	self:InitializeOptions()
	self:PopulateCdSpellsOptions()
	LibStub("AceConfig-3.0"):RegisterOptionsTable("GCD", self.options)
	self:RegisterChatCommand("gcd", self.OpenConfig)
end

function GCD:OnEnable()
	local activeSpec = GetActiveTalentGroup(false, false)
	if (activeSpec == 1 and self.db.char.talentSpecs.primary) or (activeSpec == 2 and self.db.char.talentSpecs.secondary) then
		if self.db.char.castTime then
			self:RegisterEvent("UNIT_SPELLCAST_SENT")
			self:RegisterEvent("UNIT_SPELLCAST_START")
			self:RegisterEvent("UNIT_SPELLCAST_STOP")
			self:RegisterEvent("UNIT_SPELLCAST_INTERRUPTED")
			self:RegisterEvent("UNIT_SPELLCAST_FAILED")
			self:RegisterEvent("UNIT_SPELLCAST_DELAYED")
			self:RegisterEvent("UNIT_SPELLCAST_CHANNEL_START")
			self:RegisterEvent("UNIT_SPELLCAST_CHANNEL_STOP")
			self:RegisterEvent("UNIT_SPELLCAST_CHANNEL_UPDATE")
			if self.db.char.hideCastBar then
				CastingBarFrame:UnregisterAllEvents()
				CastingBarFrame:Hide()
			end
		end
		self:RegisterEvent("ACTIONBAR_UPDATE_COOLDOWN")
		self:RegisterEvent("SPELLS_CHANGED")
		self:SPELLS_CHANGED()
	end
	self:RegisterEvent("ACTIVE_TALENT_GROUP_CHANGED")
end

function GCD:OnDisable()
	if GCDBar then GCDBar:Hide() end
	self:UnregisterAllEvents()
	if self.db.char.hideCastBar then
		CastingBarFrame:UnregisterAllEvents()
		CastingBarFrame:RegisterEvent("UNIT_SPELLCAST_START")
		CastingBarFrame:RegisterEvent("UNIT_SPELLCAST_STOP")
		CastingBarFrame:RegisterEvent("UNIT_SPELLCAST_FAILED")
		CastingBarFrame:RegisterEvent("UNIT_SPELLCAST_INTERRUPTED")
		CastingBarFrame:RegisterEvent("UNIT_SPELLCAST_DELAYED")
		CastingBarFrame:RegisterEvent("UNIT_SPELLCAST_CHANNEL_START")
		CastingBarFrame:RegisterEvent("UNIT_SPELLCAST_CHANNEL_STOP")
		CastingBarFrame:RegisterEvent("UNIT_SPELLCAST_CHANNEL_UPDATE")
	end
end

function GCD:ACTIVE_TALENT_GROUP_CHANGED()
	local activeSpec = GetActiveTalentGroup(false, false)
	if (activeSpec == 1 and self.db.char.talentSpecs.primary) or (activeSpec == 2 and self.db.char.talentSpecs.secondary) then
		self:OnEnable()
	else
		self:OnDisable()
		self:RegisterEvent("ACTIVE_TALENT_GROUP_CHANGED")
	end
end

function GCD:ApplySettings()
	if not GCDBar then
		GCDBar = createFrame('Frame')
	end
	GCDBar:Hide()
	GCDBar:ClearAllPoints()
	GCDBar:SetFrameStrata('HIGH')
	GCDBar:SetScript('OnShow', OnShow)
	GCDBar:SetScript('OnHide', OnHide)
	GCDBar:SetWidth(64)
	GCDBar:SetHeight(64)
	GCDBar.xOff = 10
	GCDBar.yOff = -10
	
	if self.db.char.ring then
		if not GCDBar.texture then
			GCDBar.texture = GCDBar:CreateTexture(nil, 'ARTWORK')
			GCDBar.texture.timer = 0;
			GCDBar.texture.hAngle = 0;
		end
		GCDBar.texture:SetTexture(self.db.profile.ring.texture) -- Spells\\AURARUNE8
		GCDBar.texture:SetVertexColor(self.db.profile.ring.color.r,self.db.profile.ring.color.g,self.db.profile.ring.color.b,self.db.profile.ring.color.a) -- 0,1,0,0.5
		GCDBar.texture:SetBlendMode('ADD')
		GCDBar.texture:SetWidth(80)
		GCDBar.texture:SetHeight(80)
		GCDBar.texture:SetPoint('CENTER', GCDBar, 'CENTER')
		local s = sin(rad(GCDBar.texture.hAngle));
		local c = cos(rad(GCDBar.texture.hAngle));
		GCDBar.texture:SetTexCoord(0.5-s, 0.5+c,
											0.5+c, 0.5+s,
											0.5-c, 0.5-s,
											0.5+s, 0.5-c);
	else
		if GCDBar.texture then
			GCDBar.texture:ClearAllPoints()
			GCDBar.texture:Hide()
		end
	end
	
	if self.db.char.spark then
		if not GCDBar.spark then
			GCDBar.spark = GCDBar:CreateTexture(nil, 'DIALOG')
		end
		GCDBar.spark:SetTexture(self.db.profile.spark.texture) -- Spells\\AURA_01
		GCDBar.spark:SetBlendMode('ADD')
		GCDBar.spark:SetVertexColor(self.db.profile.spark.color.r,self.db.profile.spark.color.g,self.db.profile.spark.color.b,self.db.profile.spark.color.a) -- 1,1,1
		GCDBar.spark:SetWidth(30)
		GCDBar.spark:SetHeight(30)
	end
	
	if self.db.char.castTime then
		if not GCDBar.castSpark then
			GCDBar.castSpark = GCDBar:CreateTexture(nil, 'DIALOG')
		end
		GCDBar.castSpark:SetTexture(self.db.profile.castTime.texture) -- Spells\\AURA_01
		GCDBar.castSpark:SetVertexColor(self.db.profile.castTime.castColor.r, self.db.profile.castTime.castColor.g, self.db.profile.castTime.castColor.b, self.db.profile.castTime.castColor.a) -- 0,1,0,1
		GCDBar.castSpark:SetBlendMode('ADD')
		GCDBar.castSpark:SetWidth(30)
		GCDBar.castSpark:SetHeight(30)
		
		if self.db.char.spellText then
			if not GCDBar.spellText then
				GCDBar.spellText = GCDBar:CreateFontString(nil, "OVERLAY")
			end
			GCDBar.spellText:ClearAllPoints()
			GCDBar.spellText:SetPoint("BOTTOM", GCDBar, "TOP", 0, 5)
			GCDBar.spellText:SetFont(media:Fetch("font", self.db.profile.castTime.spellText.font), self.db.profile.castTime.spellText.fontSize)
		end
	end
	if self.db.char.spellCds then
		for i,v in ipairs(cdFrames) do
			if not v.frame and v.spell then
				local frame = createFrame('Frame')

				frame:ClearAllPoints()
				frame:SetFrameStrata('HIGH')
				frame:SetWidth(15*frame:GetEffectiveScale())
				frame:SetHeight(15*frame:GetEffectiveScale())
				frame:SetScript('OnShow', OnShow)
				frame:SetScript('OnHide', OnHide)
				frame:Hide()
				frame.xOff = v.xOff
				frame.yOff = v.yOff
				
				frame.spell = true
				
				frame.texture = frame.texture or frame:CreateTexture(nil, 'ARTWORK')
				frame.texture:SetTexture(v.icon)
				frame.texture:SetVertexColor(1,0,0)
				frame.texture:SetWidth(15*frame:GetEffectiveScale())
				frame.texture:SetHeight(15*frame:GetEffectiveScale())
				frame.texture:SetPoint('CENTER', frame, 'CENTER')
				
				frame.cdText = frame.cdText or frame:CreateFontString()
				frame.cdText:SetFont("Fonts\\FRIZQT__.TTF", 11, "OUTLINE, MONOCHROME")
				frame.cdText:SetPoint('CENTER', frame, 'CENTER', 0, 1)
				v.frame = frame
			elseif v.frame then
				v.frame.texture:SetTexture(v.icon)
			end
		end
	end	
end

function GCD:OpenConfig()
	ACD:Open("GCD")
end

function GCD:ACTIONBAR_UPDATE_COOLDOWN()
	if spellNum then
		local start, dur = GetSpellCooldown(spellNum, BOOKTYPE_SPELL)
		if dur > 0 and dur <= 1.5 and self.db.char.spark then
			isGCD = true
			GCDBar.startTime = start
			GCDBar.duration = dur
			GCDBar:Show()
		end
		if self.db.char.spellCds then
			local _, gcdLeft = GetSpellCooldown(spellNum, BOOKTYPE_SPELL)
			for _,v in ipairs(cdFrames) do
				if v.spell then
					local start, dur = GetSpellCooldown(v.spell, BOOKTYPE_SPELL)
					if dur > gcdLeft then
						v.frame.startTime = start
						v.frame.duration = dur
						v.frame:Show()
					end
				end
			end
		end
	end
end

function GCD:UNIT_SPELLCAST_SENT(_, unit)
	if unit ~= 'player' then return end
	castSent = GetTime()*1000
end

function GCD:UNIT_SPELLCAST_START(_, unit, spell)
	if unit ~= 'player' then return end
	isCasting = true
	castInterrupted = false
	_, _, _, icon, castStartTime, castEndTime = UnitCastingInfo(unit)
	local sendLag= (castSent and castSent > 0) and GetTime()*1000-castSent or 0
	castDuration=castEndTime-castStartTime
	sendLag = sendLag > castDuration and castDuration or sendLag
	castLatency = sendLag/castDuration
	if GCDBar.spellText then
		GCDBar.spellText:SetText(spell)
	end
	GCDBar:Show()
end

function GCD:UNIT_SPELLCAST_STOP(_, unit)
	if unit ~= 'player' then return end
  isCasting = false
end

function GCD:UNIT_SPELLCAST_FAILED(event, unit)
	if unit ~= 'player' then return end
	isCasting = false
end

function GCD:UNIT_SPELLCAST_INTERRUPTED(event, unit)
	if unit ~= 'player' then return end
	isCasting = false
end

function GCD:UNIT_SPELLCAST_DELAYED(event, unit)
	if unit ~= 'player' then return end
	castInterrupted = true
	_, _, _, _, castStartTime, castEndTime = UnitCastingInfo(unit)
end

function GCD:UNIT_SPELLCAST_CHANNEL_START(event,unit)
	if unit ~= 'player' then return end
	isCasting = true
	_, _, _, _, castStartTime, castEndTime = UnitChannelInfo(unit)
	local sendLag = (castSent and castSent > 0) and GetTime() * 1000 - castSent or 0
	castDuration = castEndTime - castStartTime
	sendLag = sendLag > castDuration and castDuration or sendLag
	castLatency = sendLag / castDuration
	if GCDBar.spellText then
		GCDBar.spellText:SetText(spell)
	end
	GCDBar:Show()
end

function GCD:UNIT_SPELLCAST_CHANNEL_STOP(event,unit)
	if unit ~= 'player' then return end
	isCasting = false
end

function GCD:UNIT_SPELLCAST_CHANNEL_UPDATE(event,unit)
	if unit ~= 'player' then return end
	castInterrupted = true
	_, _, _, _, castStartTime, castEndTime = UnitChannelInfo(unit)
end

function GCD:SPELLS_CHANGED()
	local _, class = UnitClass("player")
	spellNum = GetSpellPosInSpellbook(GetSpellInfo(spells[class]))
	
	for _,v in ipairs(cdFrames) do
		if v.frame then deleteFrame(v.frame) end
	end
	cdFrames = {}
	for i,v in pairs(self.db.char.cdSpells) do
		local spell, _, icon = GetSpellInfo(v.spellID)
		tinsert(cdFrames, {['spell'] = GetSpellPosInSpellbook(spell), ['icon'] = icon, ['xOff'] = v.xOff, ['yOff'] = v.yOff})
	end
	self:ApplySettings()
end

function GCD:Unlock()
	if self.db.char.spellCds then
		local cursor = createFrame("Frame")
		cursor:SetWidth(20)
		cursor:SetHeight(20)
		cursor:SetBackdrop({bgFile = "Interface/Tooltips/UI-Tooltip-Background", 
															  edgeFile = "Interface/Tooltips/UI-Tooltip-Border", 
															  tile = true, tileSize = 16, edgeSize = 16, 
															  insets = { left = 4, right = 4, top = 4, bottom = 4 }})
		cursor:SetBackdropColor(255,0,0,1)
		cursor:SetFrameStrata('HIGH')
		cursor:SetPoint("CENTER", UIParent, "BOTTOMLEFT", (self.db.profile.position.x+10)/cursor:GetEffectiveScale(), (self.db.profile.position.y-10)/cursor:GetEffectiveScale())
		cursor:SetScript("OnMouseDown", cursor.StartMoving)
		cursor:SetScript("OnMouseUp", function(self, button)
												self:StopMovingOrSizing()
												if button == "RightButton" then
													GCD.db.profile.position.x, GCD.db.profile.position.y = (self:GetLeft() + self:GetWidth() / 2) - 10, (self:GetBottom() + self:GetHeight() / 2) + 10
													deleteFrame(self)
													if GCD.db.char.spellCds then
														for i,v in ipairs(cdFrames) do
															GCDBar:Hide()
															GCDBar:SetScript('OnShow', OnShow)
															GCDBar:SetScript('OnHide', OnHide)
															v.frame:Hide()
															v.frame:EnableMouse(false)
															v.frame:SetMovable(false)
															v.frame:SetScript("OnMouseDown", nil)
															v.frame:SetScript("OnMouseUp", nil)
															v.frame:SetScript('OnShow', OnShow)
															v.frame:SetScript('OnHide', OnHide)
															
															v.frame.xOff = v.frame:GetLeft() + v.frame:GetWidth() / 2 - (self:GetLeft() + self:GetWidth() / 2) + 10
															v.frame.yOff = v.frame:GetBottom() + v.frame:GetHeight() / 2 - (self:GetBottom() + self:GetHeight() / 2) - 10
															GCD.db.char.cdSpells[i].xOff = v.frame.xOff
															GCD.db.char.cdSpells[i].yOff = v.frame.yOff
														end
													end
												end
											end)
		cursor:SetScript("OnEnter", function(self)
												GameTooltip:SetOwner(self, "ANCHOR_CURSOR")
												GameTooltip:SetText("Right Click")
												GameTooltip:AddLine("to lock the icons and save their positions", 1, 1, 1)
												GameTooltip:Show()
											end)
		cursor:SetScript("OnLeave", function(self)
												GameTooltip:Hide()
											end)
		cursor:SetMovable(true)
		cursor:EnableMouse(true)
		cursor:Show()
		
		GCDBar:SetPoint("CENTER", cursor, "CENTER")
		GCDBar:SetScript('OnShow', nil)
		GCDBar:SetScript('OnHide', nil)
		GCDBar:Show()
		
		for i,v in ipairs(cdFrames) do
			v.frame:Hide()
			v.frame:SetScript('OnShow', nil)
			v.frame:SetScript('OnHide', nil)
			v.frame:SetPoint('CENTER', cursor, 'CENTER', (v.frame.xOff-10)/v.frame:GetEffectiveScale(), (v.frame.yOff+10)/v.frame:GetEffectiveScale())
			v.frame.texture:SetVertexColor(1.0,1.0,1.0)
			v.frame.cdText:SetText("")
			v.frame:Show()
			v.frame:SetMovable(true)
			v.frame:EnableMouse(true)
			v.frame:SetScript("OnMouseDown", v.frame.StartMoving)
			v.frame:SetScript("OnMouseUp", v.frame.StopMovingOrSizing)
		end
		ACD:Close("GCD")
		return true
	else
		return L["All unlockable modules are disabled!"]
	end
end
